<?php

namespace App\Http\Requests;

use App\Rules\Dashboard\ValidActiveAdminPassword;
use App\Rules\Dashboard\ValidActiveAdminUser;
use Illuminate\Contracts\Validation\ValidationRule;
use Illuminate\Foundation\Http\FormRequest;
use Illuminate\Validation\Rule;
use Illuminate\Support\Facades\Validator;

class LoginRequest extends FormRequest
{
  /**
   * Determine if the user is authorized to make this request.
   */
  public function authorize(): bool
  {
    return true;
  }

  /**
   * Get the validation rules that apply to the request.
   *
   * @return array<string, ValidationRule|array|string>
   */
  public function rules(): array
  {
    $usernameType = $this->usernameType($this->input('username'));

    return [
      'username' => ['required', 'exists:admins,' . $usernameType, new ValidActiveAdminUser($usernameType, $this->input('username'))],
      'password' => ['required', 'string', 'min:6', new ValidActiveAdminPassword($usernameType, $this->input('username'))]
    ];
  }

  /**
   * Custom messages for validation errors.
   *
   * @return array<string, string>
   */
  public function messages(): array
  {
    return [
      'username.required' => 'The username field is required.',
      'username.email' => 'The username must be a valid email address.',
      'username.phone' => 'The username must be a valid phone number.',
      'username.exists' => 'The selected username does not exist in our records.',
      'password.required' => 'The password field is required.',
      'password.min' => 'The password must be at least 6 characters.'
    ];
  }

  /**
   * Determine if the username is an email or mobile number.
   *
   * @param string|null $username
   * @return string
   */
  public function usernameType(string $username = null): string
  {
    if (filter_var($username, FILTER_VALIDATE_EMAIL)) {
      return 'email';
    }

    if (preg_match('/^\+?[1-9]\d{1,14}$/', $username)) {
      return 'phone';
    }

    return 'username';
  }

  /**
   * Prepare the data for validation.
   */
  protected function prepareForValidation(): void
  {
    $this->merge([
      'username' => trim($this->input('username')),
      'password' => trim($this->input('password'))
    ]);
  }

  /**
   * Get the validated data from the request.
   *
   * @param string|null $key
   * @param mixed $default
   * @return array
   */
  public function validated($key = null, $default = null): array
  {
    return array_merge(parent::validated(), [
      'username_type' => $this->usernameType($this->input('username'))
    ]);
  }
}
